<?php

/* NoPro handy-lib
 * pre-parse (JS-)script-files
 */

if (!defined('CRLF')) define('CRLF', "\r\n");

function wrap_4dbg($txt) { return preg_replace("/([}{;])/", "$1".CRLF, $txt); }

class fn_entity {
  private
    $name = '',
    $pos = 0,
    $local = false,
    $fbeg = null,
    $fend = null;
  
  public function fn_entity($n, $p, $stat=false, $fp0=null, $fp1=null) {
    $this->name = $n;
    $this->pos = $p;
    $this->local = $stat;
    $this->fbeg = $fp0;
    $this->fend = $fp1;
  }
  public function get_name() { return $this->name; }
  public function get_dpos() { return $this->pos; }
  public function get_stat() { return $this->local; }
  public function get_fpos() { return $this->fbeg; }

  public function set_stat($stat) { $this->local = $stat; }
  public function set_fpos($fp0, $fp1) { $this->fbeg = $fp0; $this->fend = $fp1; }

}

class parser {
  private $txt = '', $uri = '';
  public function get_uri() { return $this->uri; }

  public $code    = array();		// array of fragments without strings
  public $strings = array();		// array of strings
  public $cfnames = array();		// array of function names
  public $cvnames = array();		// array of var-names
  public $token   = array(		// array of unique random subst-names
			  'str' => '',
			  'fun' => '',
			  'var' => ''
			  );
  public function get_token($nam) { return $this->token[$nam]; }
  public function set_token($nam, $val) { $this->token[$nam] = $val; }

  public function parser($url, $pp=false) { $this->load($url); if($pp) $this->preparse(); }
  public function load($url) {
    $this->uri = $url;
    $this->txt = file_get_contents($url);
    if($this->txt === false) {
      $this->txt = '';
    }
    return strlen($this->txt);
  }
  public function nplpack() { return array('s' => $this->strings, 'c' => $this->code); }

  public function text_full() {
    // glue together fragments and strings
    for($txt='', $i=0; $i < count($this->strings); $i++)
      $txt .= $this->code[$i] . $this->strings[$i];
    return $txt . $this->code[$i];
  }

  public function text_subs() {
    // glue together fragments and reference to strings
    $txt = $this->get_token('str'). '='. json_encode($this->strings).';';
    for($i=0; $i < count($this->strings); $i++)
      $txt .= $this->code[$i] . $this->get_token('str') . "[$i]";
    return $txt . $this->code[$i];
  }

  public function text_4dbg() {
    // text_full() added with some crlf
    $txt = '';
    reset($this->code);
    foreach($this->strings as $s) {
      list($i, $c) = each($this->code);
      $txt .= wrap_4dbg($c) . $s;	// dont wrap strings
    }
    list($i, $c) = each($this->code);
    $txt .= wrap_4dbg($c);
    return $txt;
  }

  public function preparse($txt=NULL, $type = "JS") {
    if($txt == NULL) $txt = $this->txt;
    // remove remarks and populate arrays $code and $strings
    $this->code = $this->strings = array();

    $e = array();
    // store dbl-escapes
    for($i=0; $i < strlen($txt); $i = $p1+2) {
      $p1 = strpos($txt, "\\\\", $i);
      if($p1 === false) break;
      array_push($e, $p1);
    }

    $q1 = array('#',  '//', '/*', '/', '"', "'");
    $q2 = array("\n", "\n", '*/', '/', '"', "'");	// \n must be inside dbl-quotes
    ///////////  0     1     2     3    4    5

    // find every first q-mark and its end-delimitor
    for($bp = 0; $bp < strlen($txt); $bp = $p2) {

      // find q-start p1
      //
      // init p1 and x
      $p1 = strlen($txt);
      $x = -1;
      for($i=0; $i<count($q1); $i++) {
	if($i==0 && $type == 'JS') continue;			// skip '#' for JS
	if($i==3 && $type != 'JS') continue;			// skip '/' for non-JS (regex delimitor)
	if(($i==1 || $i==2) && $type =='SH') continue;		// skip '//' anf '/*' for SH

	$tmp = $bp-1;
	// loop for regex only
	// no need to loop for strings, because the first string-qchar cannot be escaped
	do {
	  $tmp = strpos($txt, $q1[$i], 1+$tmp);
	  if($tmp !== false && $i==3 && $tmp < $p1) {		// JS-regex /.../
	    for($tmp2 = $tmp-1; $tmp2 > 0; $tmp2--)
	      if(strstr(" \t\r\n", $txt[$tmp2]) === false) break; // first non wspace before '/'
	    if(strstr('=,:{([', $txt[$tmp2]) !== false) break;	// JS-regex found
	  }
	} while($tmp !== false && $i==3 && $tmp < $p1);		// search next '/'
	if($tmp !== false && $tmp < $p1) {			// smaller p1 found
	  $p1 = $tmp;
	  $x = $i;
	}
      }
      if($x < 0) break;	// no more q-mark found

      // we have smallest p1 now
      // find q-end p2
      $p2 = $p1 + strlen($q1[$x]) -1;
      do {
	$p2 = strpos($txt, $q2[$x], 1+$p2);
	//echo "$x: $p1 - $p2\n";
      } while($p2 !== false && $x>3 && substr($txt, $p2-1, 1) == "\\" && !in_array($p2-2, $e));
      if($p2 === false) break;
      $p2 += strlen($q2[$x]);

      // store
      array_push($this->code, substr($txt, $bp, $p1-$bp));		// code-fragment upto quote
      if($x < 3)
	array_push($this->strings, null);				// ignore remarks
      else
	array_push($this->strings, substr($txt, $p1, $p2-$p1));		// string with surrounding quotes from p1-p2

      if(DBGverbous && $x==3) echo 'JS-regex: '. substr($txt, $p1, $p2-$p1) .CRLF;

    }
    if($bp < strlen($txt))	// final code-fragment
      array_push($this->code, substr($txt, $bp));

    $this->cpack();
  }

  public function eoterm($txt, $s0, $s1, $off=0) {
    // returns strpos of end of term () {}
    $p1 = strpos($txt, $s0, $off);	// begin of term
    if($p1 === false) return false;	// not found
    $p2 = strpos($txt, $s1, $p1);	// perhaps end of term
    if($p2 === false) return false;	// none found

    $p3 = strpos($txt, $s0, $p1+1);	// next begin of term
    $n = ($p3 && $p3 < $p2) ? 2 : 0;	// init depth of recursion
    for(;$n;) {
      $p2 = strpos($txt, $s1, $p3+1);	// end of recursive term
      if($p2 === false) return false;	// none found
      $p3 = strpos($txt, $s0, $p3+1);	// begin of next recursive term
      if($p3 && $p3 < $p2)		// one more recursion found
	$n++;
      else {				// one recursion ended
	$p3 = $p2;
	$n--;
      }
    }
    return $p2; // return strpos of end of term
  }

  private function nplmini($arr) {
    // shrink code
    $r1 = array(
		"/\r/",
		"/\n/",
		"/\s/",	// "/\h/",	// h-whitespace
		"/  +/",
		"/ *([,;:=+\-*\/|&^\[\]{}()<>?]) */",	/// superfluous spaces
		"/;+}/",	// javascript doesnt need ';' at end of block
		"/ +$/",	// ltrim
		"/^ +/",	// rtrim
		);
    $s2 = array(
		' ',	// cr
		' ',	// lf
		' ',	// h-whitespace
		' ',	// multiple spaces
		'${1}',	// superfluous spaces
		'}',	// no semicolon before '}'
		'',	// trim
		'',	// trim
		);
    return preg_replace($r1, $s2, $arr);
  }

  private function cpack() {
    #print_r($this->code); print_r($this->strings); exit;

    // pack code by removing former comments
    // (concat code-fragments if s[] == null)
    $c0 = $this->code;
    $s0 = $this->strings;
    $c1 = $s1 = array();
    $cc = '';
    reset($s0);
    foreach($c0 as $cs) {
      $cc .= " ".$cs;
      list($sk, $sv) = each($s0);
      if($sv != null) {
	array_push($s1, $sv);
	array_push($c1, $this->nplmini(substr($cc, 1)));
	$cc = '';
      }
    }
    array_push($c1, $this->nplmini(substr($cc, 1)));
    //$this->strings = $s1; $this->code = $c1; return;

    // perform a tiny syntax-check
    //
    // 1. add semicolon
    // 1a) x=function(){};
    // 1b) {for();}
    // 1c) ={};

    // init: concat code without strings and
    // store total len after each fragment
    $JS = '';			// glued fragments
    $parr = array();		// insert-array pos => val
    $l=0; $jp=array($l);	// total len
    foreach($c1 as $c) {
      $JS .= $c;
      $l += strlen($c);
      array_push($jp, $l);
    }

    // 1. add semicolon // 1a) x=function(){};
    // syntax error: x=new function(){};
    for($p1 = strpos($JS, '=function('); $p1 !== false; $p1 = strpos($JS, '=function(', $p1+1)) {
      $p2 = $this->eoterm($JS, '(', ')', $p1); if($p2 === false) break;
      $p2 = $this->eoterm($JS, '{', '}', $p2); if($p2 === false) break;
      if(++$p2+1 < $l && (strstr(';,})(:', $JS[$p2]) === false))	// ;,})(:
	$parr[$p2] = ';';
    }

    // 1. add semicolon // 1b) {for();}	(found in bigint.js)
    for($p1 = strpos($JS, 'for('); $p1 !== false; $p1 = strpos($JS, 'for(', $p1+1)) {
      $p2 = $this->eoterm($JS, '(', ')', $p1); if($p2 === false) break;
      if(++$p2+1 < $l && $JS[$p2] == '}')
	$parr[$p2] = DBGverbous ? ';;' : ';';
    }

    // 1. add semicolon // 1c) ={};
    for($p1 = strpos($JS, '={'); $p1 !== false; $p1 = strpos($JS, '={', $p1+1)) {
      $p2 = $this->eoterm($JS, '{', '}', $p1); if($p2 === false) break;
      if(++$p2+1 < $l && (strstr(';,})', $JS[$p2]) === false))	// ;,})
	$parr[$p2] = DBGverbous ? ';;;' : ';';
    }

    #print_r($parr); print_r(array_reverse($parr, true)); exit;
    // exec inserts last -> first
    foreach(array_reverse($parr, true) as $p => $s) {
      if(DBGverbous) echo CRLF. "jsi '$s' @$p: ". substr($JS, $p-5, 5) .' # '. substr($JS, $p, 5) .CRLF;	// show 5 context-chars
      jsinsert($c1, $jp, $p, $s);
    }

    $this->strings = $s1;
    $this->code = $c1;
  }

  public function nuniq($v0) {
    // new uniq name not in code, vars or token
    for($uniq = $v0;;) {
      $uniq .= chr(96+mt_rand(1, 26));
      if(count(preg_grep("/$uniq/", $this->token)) > 0) continue;
      if(count(preg_grep("/$uniq/", $this->code)) > 0) continue;
      /**
      if(count(preg_grep("/$uniq/", $this->cfnames)) > 0) continue;
      if(count(preg_grep("/$uniq/", $this->cvnames)) > 0) continue;
      /**/
      break;
    }
    return $uniq;
  }

  public function parse_js ($vlen=4, $verbose=false) {
    // populate arrays $cfnames and $cvnames
    // vlen ~ minimal length for var-names

    $rgx = array(
		 'fun' => "/(^|[;}])function /",	// named fn, w-space were trimmed before for anon-fn
		 'var' => "/(^|[];}{)])var /",
		 );

    // get fnames and whole var-lines
    $txt = implode("''", $this->code);
    foreach($rgx as $k => $v) {
      $map[$k] = array();
      $n = preg_match_all($v, $txt, $tmp);
      for($bp0=$i=0; $i < $n; $i++) {
	$bp0 = strpos($txt, $tmp[0][$i], $bp0) + strlen($tmp[0][$i]);
	if($k == 'var') {
	  $bp1 = strpos($txt, ';', $bp0);
	  $bp2 = strpos($txt, '}', $bp0);
	  if($bp2 && $bp2 < $bp1 && $bp2) {
	    $bp3 = strpos($txt, '{', $bp0);
	    if($bp3 === false || $bp2 < $bp3)
	      $bp1 = $bp2;
	  }
	} else if($k == 'fun') {
	  $bp1 = strpos($txt, '(', $bp0);
	  // fn-boundaries
	  $bp2 = $this->eoterm($txt, '(', ')', $bp1) + 1;
	  $bp3 = $this->eoterm($txt, '{', '}', $bp2);
	}

	if($k == 'fun') {
	  $map[$k][] = $nmo = new fn_entity(substr($txt, $bp0, $bp1 - $bp0), $bp0);
	  $nmo->set_fpos($bp2, $bp3);
	} else {
	  $map[$k][] = substr($txt, $bp0, $bp1 - $bp0);
	}

      }
    }
    $fnames = array();
    foreach($map['fun'] as $v) {
      $fnames[] = $v->get_name();
    }
    $this->cfnames = array_unique($fnames);
    if($verbose == 2) {
      foreach($rgx as $k => $v) { echo "$k: "; print_r($map[$k]); } #exit;
    }

    // analyze varlines for varnames
    $avn = array();
    foreach($map['var'] as $k => $v) {
      if($verbose)
	echo "var $k: $v" .CRLF;
      for($bp0=0, $bp1=strcspn($v, '=,.', $bp0); $bp0 < strlen($v); $bp0=$bp1+1, $bp1=$bp0+strcspn($v, '=,.', $bp0)) {
	$vn = substr($v, $bp0, $bp1-$bp0);
	if($verbose)
	  echo "  $vn".CRLF;
	if($bp0+$vlen <= $bp1)
	  array_push($avn, $vn);

	if($v[$bp1] == '.')		// object
	  $bp1 += strcspn($v, '=,', $bp1);

	while($v[$bp1] == '=' || $v[$bp1] == ']' || $v[$bp1] == ')' || $v[$bp1] == '}') {
	  $bp1 += strcspn($v, ',[({', $bp1);
	  #if($bp1 >= strlen($v)) break;
	  if($v[$bp1] == '[')
	    $bp1 = $this->eoterm($v, '[', ']', $bp1);
	  else if($v[$bp1] == '(')
	    $bp1 = $this->eoterm($v, '(', ')', $bp1);
	  else if($v[$bp1] == '{')
	    $bp1 = $this->eoterm($v, '{', '}', $bp1);
	  else if($verbose > 1)
	    echo "bp=$bp1/". strlen($v) ." '". $v[$bp1] ."'" .CRLF;
	  if($bp1 === false) $bp1 = strlen($v);
	}
      }
    }
    $this->cvnames = array_unique($avn);
  }

}

// find idx in array: a(i) < x <= a(i+a)
function heron($a, $x) {
  // intervall-schachtelung durch Halbieren
  for($idx = (int)(count($a)/2 + .5), $n=$idx/2; $n > .5; $n/=2) {
    if($x < $a[$idx])
      $idx -= (int)($n+.5);
    else if($x > $a[$idx+1])
      $idx += (int)($n+.5);
    else
      break;
  }
  // korrigiere wenn knapp daneben
  if($x < $a[$idx]) $idx--;
  if($x > $a[$idx+1]) $idx++;
  if($a[$idx] < $x && $x <= $a[$idx+1])
    return $idx;

  // grosse schleife weil total daneben
  if(DBGverbous) echo "heron 1: ". $a[$idx] .' # '. $x .' # '. $a[$idx+1] .CRLF;
  for($idx=0; $idx < count($a)-1; $idx++) {
    if($x <= $a[$idx+1]) break;
  }
  if(DBGverbous) echo "heron 2: ". $a[$idx] .' < '. $x .' <= '. $a[$idx+1] .CRLF;

  return $idx;
}

function jsinsert(&$carr, $larr, $p, $s) {
  // insert $s in $carr at total strpos $p
  // $larr is array of total len after each fragment

  #echo $p .': insert('.$s,')' .CRLF;

  $i = heron($larr, $p);	// get idx for $carr
  $x = $p - $larr[$i];		// relative strpos in that fragment
  if(DBGverbous) echo $larr[$i]. " < $p ~ $i, $x: ". substr($carr[$i], $x-5, 5) .'  ;  '. substr($carr[$i], $x, 5) .CRLF;	// show 5 context-chars

  $s1 = substr($carr[$i], 0, $x);
  $s2 = substr($carr[$i], $x);
  $carr[$i] = $s1 .$s. $s2;
  if(DBGverbous) echo substr($carr[$i], $x-3, 7) .CRLF;

}

?>
